'' =================================================================================================
''
''   File....... jm_file_parsing_demo.spin
''   Purpose....  
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2017 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 10 JUN 2017 
''
'' =================================================================================================


con { timing }

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' use 5MHz crystal
                                                                 
  CLK_FREQ = _xinfreq * (_clkmode >> 6)                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us
                                                                  
                                                                  
con { pab io pins }                                               
                                                                  
  RX1     = 31                                                  ' programming / terminal
  TX1     = 30                                      
                                                    
  SDA1    = 29                                                  ' eeprom / i2c
  SCL1    = 28                                                   
                                                                 
  AUD_RT  = 27                                                  ' audio (d/a) outputs
  AUD_LF  = 26                                                   
                                                                 
  LED_27  = 27                                                  ' leds (on audio outputs)
  LED_26  = 26                                                   
                                                                 
  SD_CS   = 25                                                  ' usd card
  SD_DI   = 24                                                   
  SD_CLK  = 23                                                   
  SD_DO   = 22                                                   
                                                                 
  ADC_CS  = 21                                                  ' adc input
  ADC_CLK = 20                                                   
  ADC_DO  = 19                                                   
  ADC_DI  = 18                                                   
                                                                 
  SVO_6   = 17                                                  ' servo headers
  SVO_5   = 16                                                   
  SVO_4   = 15                                                   
  SVO_3   = 14                                                   
  SVO_2   = 13                                                   
  SVO_1   = 12                                                   

                                                       
obj                                                              
                                                                 
' main                                                          ' * master spin cog
  time   : "jm_time_80"                                         '   delays/timing
  io     : "jm_io"                                              '   basic io
  parser : "jm_parser"                                          '   text parsing
  str    : "jm_strings"                                         '   strings support
  term   : "jm_fullduplexserial"                                ' * serial for terminal
  sd     : "fsrw"                                               ' * uSD file IO
                                                                 
' * uses cog when loaded                                         
                                                                 
                                                                 
var

  byte  filename[13]
  byte  option[3]
                                                              
                                                                 
pub main | idx                                                        
                                                                 
  setup

  term.rxflush
  term.rx                                                       ' open PST, enable, press a key
  term.tx(term#CLS)

  if (read_configuration)
    term.str(string("Filename... "))
    term.str(@filename)
    term.tx(13)

    repeat idx from 0 to 2
      term.str(string("Option "))
      term.tx("0"+idx)
      term.str(string("... "))
      term.dec(option[idx])
      term.tx(13)
       
  repeat                                                         
    waitcnt(0)


dat

  CfgFile               byte    "startup.txt", 0


pub read_configuration | check

  ifnot (has_file(@CfgFile))
    term.str(string("Error: Cannot open STARTUP.TXT", 13))
    return false

  open_file(@CfgFile, "r")

  repeat
    bytefill(@linebuf, 0, 80+1)                                 ' clear for full line
    check := rd_line(@linebuf, 80)                              ' read line from file
    if (check =< 0)                                             ' if end of file
      quit
    else
      str.replace(39, 0, @linebuf)                              ' remove comments (single quote)   
      str.replace(13, 0, @linebuf)                              ' remove extra CRs
      str.trim(@linebuf)                                        ' clean-up
      if (strsize(@linebuf))                                    ' if we have a line
        check := parser.enqueue_str(@linebuf)                   ' send to parser     
        if (check)                                              ' have valid tokens?
          process_cmd                                          
          parser.reset                                          ' reset parser for next command     
                                                               
  close_file

  return true                                                   ' found and read file                                       


dat { configure parser characters and tokens }

  CHAR_SET              byte    "$%+-_"
                        byte    ".0123456789"
                        byte    "ABCDEFGHIJKLMNOPQRSTUVWXYZ", 0   
  
  TOKEN_LIST            byte    "FILENAME", 0                   ' FILENAME string
                        byte    "OPTION", 0                     ' OPTION # value

                        
con { enumerated token indexes }

  #0, T_FILENAME, T_OPTION, TOKEN_COUNT 
  

pub process_cmd | tidx

  tidx := parser.get_token_id(parser.token_addr(0))              ' get index of command token
  
  case tidx
    T_FILENAME : cmd_filename
    T_OPTION   : cmd_option


pub cmd_filename | tc, p_name

  tc := parser.token_count
  if (tc <> 2)                                                  ' syntax = FILENAME string
    return

  p_name := parser.token_addr(1)                                ' get address of name
  bytefill(@filename, 0, 13)                                    ' reset
  bytemove(@filename, p_name, strsize(p_name) <# 12)            ' copy token to filename - 12 chars max  

    
pub cmd_option | tc, idx, value

  tc := parser.token_count
  if (tc <> 3)                                                  ' syntax = OPTION # byte_value
    return

  idx := parser.token_value(1)

  if ((idx < 0) or (idx > 2))                                   ' valid option # 
    return

  option[idx] := 0 #> parser.token_value(2) <# 255              ' set option
                                                       
                                                                 
pub setup                                                        
                                                                 
  time.start                                                    ' setup for timing/delays
                                                                 
  io.start(0, 0)                                                ' clear pins (this cog)

  parser.start(@CHAR_SET, @TOKEN_LIST, TOKEN_COUNT, false)      ' setup parsing engine         
                                                              
  term.start(RX1, TX1, %0000, 115_200)                          ' start serial for terminal *
                                                                 
  mount_sd                                                      ' start sd access *


con                                                              
                                                                 
  { ----------------- }                                          
  {  F I L E   I / O  }                                          
  { ----------------- }                                          
                                                                 
                                                                 
var                                                              
                                                                 
  long  mounted                                                  
                                                                 
  byte  linebuf[80 + 1]                                         ' line buffer
                                                                 
                                                                 
pub mount_sd | check                                             
                                                                 
'' Attempt to mount SD card                                      
'' -- returns true if SD card successfully mounted

  check := \sd.mount_explicit(SD_DO, SD_CLK, SD_DI, SD_CS)

  mounted := (check == 0)   

  return mounted


pub dismount_sd

'' Unmounts SD card

  \sd.unmount
  mounted := false


pub has_file(p_str) | check

'' Returns true if file is on SD card

  check := \sd.popen(p_str, "r")
  if (check == 0)
    return true
  else
    return false 


pub open_file(p_str, mode) | check

'' Attempt to open file
'' -- p_str is pointer to filename (z-string)
'' -- mode is [lower-case] single character, (e.g., "r" for read)
'' -- sd card must be mounted before open

  if (mounted)                                                   
    check := \sd.popen(p_str, mode)                             ' attempt to open
    if (check == 0)                                              
      if (mode == "r")                                           
        \sd.seek(0)                                             ' force sector read 
      return true                                                
    else                                                         
      return false
  else
    return false


pub close_file

'' Closes open file

  \sd.pclose


pub rd_line(p_str, n) | c, len

'' Reads line of text from open text file
'' -- terminated by LF or EOF
'' -- p_str is address of string buffer
'' -- n is maximum number of characters to read from line

  len := 0                                                      ' index into string
                                                                 
  repeat n                                                       
    c := \sd.pgetc                                              ' get a character
    if (c < 0)                                                  ' end of file
      if (len == 0)                                             ' if nothing
        return -1                                               ' return empty
      else                                                       
        quit                                                     
                                                                 
    if (c <> 10)                                                ' if not LF char
      byte[p_str++] := c                                        '  add to buffer
      len++                                                      
    else                                                        ' if LF
      quit                                                      '  we're done
                                                                 
  byte[p_str] := 0                                              ' terminate end of line
                                                                 
  return len
  

pub wr_line(p_str, usecr, uself)

'' Writes z-string to open text file
'' -- set usecr to true to add CR
'' -- set uself to true to add LF

  \sd.pputs(p_str)
  if (usecr)
    \sd.pputc(13)
  if (uself)
    \sd.pputc(10)


pub rd_buf(p_buf, n) | check

'' Reads n bytes from file to buffer at p_buf
'' -- returns number of bytes read

  return \sd.pread(p_buf, n)


pub wr_buf(p_buf, n) | check   

'' Writes n bytes to file from buffer at p_buf
'' -- returns number of bytes written

  return \sd.pwrite(p_buf, n)


pub directory | fcount, check, idx, p_fname, fsize

'' Show directory (root only) of uSD card

  sd.opendir                                                    ' setup for .nextfile
  fcount := 0                                                   ' reset files count
                                                                 
  term.str(string("Directory", term#CR, term#CR))                
                                                                 
  repeat                                                         
    bytefill(@filename, 0, 13)                                   
    check := \sd.nextfile(@filename)                            ' get next file in root
    if (check == 0)                                             ' if found
      fcount += 1                                               '  increment file count
      term.tx(" ")
      term.str(@filename)                                       '  show name
      term.tx(13)                                           
    else                                                        ' else
      quit                                                      '  done
                                                                 
  if (fcount > 0)                                                
    term.tx(13)
    term.dec(fcount)
    term.str(string(" file(s) found."))
  else
    term.str(string("No files found."))

  sd.pclose


pub show_file(p_file) | len

'' Display contents of text file

  ifnot (mounted)
    term.str(string("SD not mounted.", 13))
    return

  if (open_file(p_file, "r"))
    repeat
      bytefill(@linebuf, 0, 81)
      len := rd_line(@linebuf, 80)
      if (len > 0)
        term.str(@linebuf)
        time.pause(5)
      else
        close_file
        quit 
    

dat { license }

{{

  Copyright (c) 2017 Jon McPhalen  

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}